from __future__ import annotations

from typing import BinaryIO

from dissect.esedb.tools.sru import SRU


def test_sru(sru_db: BinaryIO) -> None:
    db = SRU(sru_db)

    records = list(db.entries())
    assert len(records) == 220



from __future__ import annotations

import struct
import uuid
from functools import cached_property
from typing import TYPE_CHECKING

from dissect.esedb.c_esedb import CODEPAGE, IDBFLAG, IDXFLAG, JET_coltyp, RecordValue
from dissect.esedb.cursor import Cursor
from dissect.esedb.lcmapstring import map_string

if TYPE_CHECKING:
    from dissect.esedb.page import Node, Page
    from dissect.esedb.record import Record
    from dissect.esedb.table import Column, Table


JET_cbKeyMost_OLD = 255


class Index:
    """Represents an index on a table.

    This is still very much WIP but works for basic indexes.
    For example, none of the special flags are currently implemented.

    Args:
        table: The table this index is from.
        record: The record in the catalog for this index.
    """

    def __init__(self, table: Table, record: Record | None = None):
        self.table = table
        self.record = record
        self.esedb = table.esedb

        self.name = record.get("Name")
        flags = record.get("Flags")
        self.idb_flags = IDBFLAG(flags & 0xFFFF)
        self.idx_flags = IDXFLAG(flags >> 16)
        self._key_most = record.get("KeyMost") or JET_cbKeyMost_OLD
        self._var_seg_mac = record.get("VarSegMac") or self._key_most

    def __repr__(self) -> str:
        return f"<Index name={self.name!r}>"

    @property
    def is_primary(self) -> bool:
        return bool(self.idb_flags & IDBFLAG.Primary)

    @cached_property
    def root(self) -> Page:
        """Return the root page of this index."""
        return self.esedb.page(self.record.get("ColtypOrPgnoFDP"))

    @cached_property
    def column_ids(self) -> list[int]:
        """Return a list of column IDs that are used in this index."""
        column_ids = []
        key_field_ids = self.record.get("KeyFldIDs")
        if len(key_field_ids) % 4 == 0:
            for i in range(0, len(key_field_ids), 4):
                _, column_identifier = struct.unpack("<HH", key_field_ids[i : i + 4])
                column_ids.append(column_identifier)
        return column_ids

    @cached_property
    def columns(self) -> list[Column]:
        """Return a list of all columns that are used in this index."""
        return [self.table._column_id_map[cid] for cid in self.column_ids]

    def cursor(self) -> Cursor:
        """Create a new cursor for this index."""
        return Cursor(self)

    def search(self, **kwargs) -> Record:
        """Search the index for the requested values.

        Args:
            **kwargs: The columns and values to search for.

        Returns:
            A :class:`~dissect.esedb.record.Record` object of the found record.
        """
        return self.cursor().search(**kwargs)

    def search_key(self, key: bytes) -> Node:
        """Search the index for a specific ``key``.

        Args:
            key: The key to search for.

        Returns:
            A :class:`~dissect.esedb.page.Node` object of the found node.
        """
        return self.cursor().search_key(key)

    def key_from_record(self, record: Record) -> bytes:
        """Generate a key for this index from a record.

        Args:
            record: The record to generate a key for.
        """
        values = {c.name: record[c.name] for c in self.columns}
        return self.make_key(values)

    def make_key(self, values: dict[str, RecordValue]) -> bytes:
        """Generate a key out of the given values.

        Args:
            values: A map of the column names and values to generate a key for.
        """
        key_buf = []
        key_remaining = self._key_most

        for column in self.columns:
            if column.name not in values:
                break

            key_part = encode_key(self, column, values[column.name], self._var_seg_mac)
            key_buf.append(key_part)
            key_remaining -= len(key_part)

            if key_remaining <= 0:
                break

        key = b"".join(key_buf)
        if key_remaining < 0:
            key = key[: self._key_most]
        return key


bPrefixNull = 0x00
bPrefixZeroLength = 0x40
bPrefixNullHigh = 0xC0
bPrefixData = 0x7F
bSentinel = 0xFF


def encode_key(index: Index, column: Column, value: RecordValue, max_size: int) -> bytes:
    """Encode various values into their normalized index key form.

    Args:
        column: The column of the value to encode.
        value: The value that needs encoding.
        max_size: The maximum key segment size.
    """
    if value is None:
        return bytes([bPrefixNull])

    # All keys with data are prefixed with 0x7f (bPrefixData)
    # There are other prefixes but we don't support those yet
    key = bytearray([bPrefixData])

    if column.type == JET_coltyp.Bit:
        key.append(0xFF if value else 0x00)

    elif column.type == JET_coltyp.UnsignedByte:
        key.append(value)

    elif column.type == JET_coltyp.Short:
        # Signed integers have their MSB bit flipped
        key += struct.pack(">H", (value ^ (1 << 15)) & 0xFFFF)

    elif column.type == JET_coltyp.Long:
        key += struct.pack(">I", (value ^ (1 << 31)) & 0xFFFFFFFF)

    elif column.type in (JET_coltyp.Currency, JET_coltyp.LongLong):
        key += struct.pack(">Q", (value ^ (1 << 63)) & 0xFFFFFFFFFFFFFFFF)

    elif column.type == JET_coltyp.IEEESingle:
        value = struct.unpack("<I", struct.pack("<f", value))[0]

        value = _flip_bits(value, 32)
        key += struct.pack(">I", value)

    elif column.type in (JET_coltyp.IEEEDouble, JET_coltyp.DateTime):
        if column.type == JET_coltyp.IEEEDouble:
            value = struct.unpack("<Q", struct.pack("<d", value))[0]

        value = _flip_bits(value, 64)
        key += struct.pack(">Q", value)

    elif column.is_binary:
        key += _encode_binary(column, value, max_size)

    elif column.is_text:
        key += _encode_text(index, column, value, max_size)

    elif column.type == JET_coltyp.UnsignedLong:
        # Unsigned variants are added as is
        key += struct.pack(">I", value)

    elif column.type == JET_coltyp.GUID:
        key += _encode_guid(value)

    elif column.type == JET_coltyp.UnsignedShort:
        key += struct.pack(">H", value)

    return bytes(key)


def _encode_binary(column: Column, value: bytes, max_size: int) -> bytes:
    cbFLDBinaryChunk = 0x08
    cbFLDBinaryChunkNormalized = 0x09

    key = bytearray()

    if not len(value):
        # Empty values (but non null)
        key = bytes([bPrefixZeroLength])
    elif column.is_fixed:
        # Fixed size binary values are added as is
        if len(value) + 1 > max_size:
            value = value[:max_size]
        key += value
    else:
        # Otherwise added as chunks of 8
        num_chunks = (len(value) + 7) // 8
        # Each chunk has 1 header byte, and the key has a header byte
        key_size = (num_chunks * 9) + 1

        normalized_all = True
        if key_size > max_size:
            key_size = max_size
            normalized_all = False

        key_remaining = key_size - 1

        value_offset = 0
        value_remaining = len(value)
        while key_remaining >= 9:
            chunk = value[value_offset : value_offset + 8]
            key += chunk

            if value_remaining <= 8:
                # Last chunk
                if value_remaining == 8:
                    key.append(cbFLDBinaryChunk if normalized_all else cbFLDBinaryChunkNormalized)
                else:
                    # Pad to 8 bytes
                    key.extend([0] * (8 - len(chunk)))
                    key.append(len(chunk))
            else:
                key.append(cbFLDBinaryChunkNormalized)
                value_offset += 8
                value_remaining -= 8

            key_remaining -= 9

        if key_remaining:
            if value_remaining >= key_remaining:
                key += value[value_offset : value_offset + key_remaining]
            else:
                key += value[value_offset : value_offset + value_remaining]
                key.extend([0] * (key_remaining - value_remaining))

    return bytes(key)


def _encode_text(index: Index, column: Column, value: str, max_size: int) -> bytes:
    key = bytearray()

    if not len(value):
        # Empty values (but non null) are indicated with 0x40
        key = bytes([bPrefixZeroLength])
    elif column.encoding in (CODEPAGE.ASCII, CODEPAGE.WESTERN):
        # ASCII strings == uppercase
        if len(value) + 1 > max_size:
            value = value[:max_size]
        key += value.upper().encode()
        key.append(0)
    else:
        # Unicode strings == LCMapStringW
        flags = index.record.get("LCMapFlags")
        locale = index.record.get("LocaleName").decode("utf-16-le")
        segment = map_string(value, flags, locale)
        key += segment[:max_size]

    return bytes(key)


def _encode_guid(value: str | uuid.UUID) -> bytes:
    if isinstance(value, str):
        value = uuid.UUID(value)
    guid_bytes = value.bytes_le

    return guid_bytes[-6:] + guid_bytes[-8:-6] + guid_bytes[-10:-8] + guid_bytes[-12:-10] + guid_bytes[:-12]


def _flip_bits(value: int, size: int) -> int:
    if value & (1 << (size - 1)):
        # If the high bit is set, all bits are flipped
        return ~value & ((1 << size) - 1)
    # Otherwise only the high bit is flipped
    return value ^ (1 << (size - 1))



from unittest.mock import Mock

from dissect.esedb.c_esedb import c_esedb
from dissect.esedb.page import Page


def test_tag_state() -> None:
    mock_esedb = Mock()
    mock_esedb.has_small_pages = True

    mock_header = c_esedb.PGHDR(
        itagState=0x2069,
    ).dumps()

    page = Page(mock_esedb, 0, mock_header)
    assert page.tag_reserved == 2
    assert page.tag_count == 0x69

    mock_header = c_esedb.PGHDR(
        itagState=0x0069,
    ).dumps()

    page = Page(mock_esedb, 0, mock_header)
    assert page.tag_reserved == 1
    assert page.tag_count == 0x69



from __future__ import annotations

import gzip
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO

import pytest

if TYPE_CHECKING:
    from collections.abc import Iterator


def absolute_path(filename: str) -> Path:
    return Path(__file__).parent / filename


def open_file(name: str, mode: str = "rb") -> Iterator[BinaryIO]:
    with absolute_path(name).open(mode) as f:
        yield f


def open_file_gz(name: str, mode: str = "rb") -> Iterator[BinaryIO]:
    with gzip.GzipFile(absolute_path(name), mode) as f:
        yield f


@pytest.fixture
def basic_db() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/basic.edb.gz")


@pytest.fixture
def binary_db() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/binary.edb.gz")


@pytest.fixture
def text_db() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/text.edb.gz")


@pytest.fixture
def multi_db() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/multi.edb.gz")


@pytest.fixture
def default_db() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/default.edb.gz")


@pytest.fixture
def index_db() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/index.edb.gz")


@pytest.fixture
def large_db() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/large.edb.gz")


@pytest.fixture
def sru_db() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/SRUDB.dat.gz")


@pytest.fixture
def ual_db() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/Current.mdb.gz")


@pytest.fixture
def windows_search_db() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/Windows.edb.gz")



from __future__ import annotations

import struct
from functools import cached_property
from typing import TYPE_CHECKING, Any

from dissect.esedb import compression
from dissect.esedb.btree import BTree
from dissect.esedb.c_esedb import (
    CODEPAGE,
    COLUMN_TYPE_MAP,
    SYSOBJ,
    ColumnType,
    JET_coltyp,
    RecordValue,
)
from dissect.esedb.exceptions import NoNeighbourPageError
from dissect.esedb.index import Index
from dissect.esedb.record import Record

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.esedb.cursor import Cursor
    from dissect.esedb.esedb import EseDB
    from dissect.esedb.page import Page


class Table:
    """Represents a table in an ESE database.

    Contains all the relevant metadata of the table, as well as all columns and indices that have been added
    by the catalog.

    Args:
        esedb: An instance of :class:`~dissect.esedb.esedb.EseDB`.
        name: The table name.
        root_page: The root page of the table.
        columns: A list of :class:`Column` for this table.
        indexes: A list of :class:`~dissect.esedb.index.Index` for this table.
        record: The :class:`~dissect.esedb.record.Record` of this table from the catalog table.
    """

    def __init__(
        self,
        esedb: EseDB,
        name: str,
        root_page: int,
        columns: list[Column] | None = None,
        indexes: list[Index] | None = None,
        record: Record = None,
    ):
        self.esedb = esedb

        self._column_name_map: dict[str, Column] = {}
        self._column_id_map: dict[int, Column] = {}
        self._index_name_map: dict[str, Index] = {}

        self.name = name
        self.root_page = root_page
        self.columns: list[Column] = []
        self.indexes: list[Index] = []

        # Set by the catalog during parsing
        self._long_value_record: Record = None
        self._long_callback_record: Record = None

        self._fixed_value_offset = 0

        columns = columns or []
        for column in columns:
            self._add_column(column)

        indexes = indexes or []
        for index in indexes:
            self._add_index(index)

        self.record = record

    def __repr__(self) -> str:
        return f"<Table name={self.name!r}>"

    @cached_property
    def root(self) -> Page:
        """Return the root page of the table."""
        return self.esedb.page(self.root_page)

    @cached_property
    def lv_page(self) -> Page:
        """Return the long value page of the table.

        Raises:
            TypeError: If the table has no long values.
        """
        if not self._long_value_record:
            raise TypeError(f"Table has no long values: {self.name}")

        return self.esedb.page(self._long_value_record.get("ColtypOrPgnoFDP"))

    def column(self, name: str) -> Column:
        """Return the column with the given name.

        Args:
            name: The name of the column to return.

        Raises:
            KeyError: If no column with the given name exists.
        """
        try:
            return self._column_name_map[name]
        except KeyError:
            raise KeyError(f"No column with this name in table {self.name}: {name}")

    @property
    def column_names(self) -> list[str]:
        """Return a list of all the column names."""
        return list(self._column_name_map.keys())

    @property
    def primary_index(self) -> Index | None:
        # It's generally the first index, but loop just in case
        for index in self.indexes:
            if index.is_primary:
                return index
        return None

    def cursor(self) -> Cursor | None:
        """Create a new cursor for this table."""
        primary_idx = self.primary_index
        if primary_idx:
            return primary_idx.cursor()
        return None

    def index(self, name: str) -> Index:
        """Return the index with the given ``name``.

        Args:
            name: The name of the index to return.

        Raises:
            KeyError: If no index with the given name exists.
        """
        try:
            return self._index_name_map[name]
        except KeyError:
            raise KeyError(f"No index with this name in table {self.name}: {name}")

    def find_index(self, column_names: list[str]) -> Index | None:
        """Find the most suitable index to search for the given columns.

        Args:
            column_names: A list of column names to find the best index for.
        """
        best_match = 0
        best_index = None
        for index in self.indexes:
            # We want to find the index that has the most matching columns in the order they are indexed
            i = 0
            for column in index.columns:
                if column.name not in column_names:
                    break
                i += 1

            if i > best_match:
                best_index = index
                best_match = i

        return best_index

    def search(self, **kwargs: RecordValue) -> Record | None:
        """Search for a record in the table.

        Args:
            **kwargs: The columns and values to search for.

        Returns:
            The first record that matches the search criteria, or ``None`` if no record was found.
        """
        return self.cursor().search(**kwargs)

    def records(self) -> Iterator[Record]:
        """Return an iterator of all the records of the table."""
        for node in self.root.iter_leaf_nodes():
            yield Record(self, node)

    def get_long_value(self, key: bytes) -> bytes:
        """Retrieve a value from the long value page of the table.

        Args:
            key: The lookup key for the long value.
        """
        rkey = key[::-1]
        btree = BTree(self.esedb, self.lv_page)
        header = btree.search(rkey)

        _, size = struct.unpack("<2I", header.data)
        chunks = []
        chunk_offsets = []

        while True:
            try:
                node = btree.next()
                if not node.key.startswith(rkey):
                    break
            except NoNeighbourPageError:
                break

            chunks.append(node.data)

            chunk_offset = struct.unpack(">I", node.key[-4:])[0]
            chunk_offsets.append(chunk_offset)

        chunk_offsets.append(size)

        buf = []
        chunk_offset = 0
        for chunk, next_chunk_offset in zip(chunks, chunk_offsets[1:], strict=False):
            # Chunk sizes should be used to determine if a chunk is compressed
            if len(chunk) != next_chunk_offset - chunk_offset:
                chunk = compression.decompress(chunk)
            buf.append(chunk)
            chunk_offset += next_chunk_offset

        return b"".join(buf)

    def _add_column(self, column: Column) -> None:
        """Add a column to the table."""
        self.columns.append(column)
        self._column_name_map[column.name] = column
        self._column_id_map[column.identifier] = column

        # Precalculate fixed value offsets
        if column.identifier < 128:
            column._offset = self._fixed_value_offset
            self._fixed_value_offset += column.size

    def _add_index(self, index: Index) -> None:
        """Add an index to the table."""
        self.indexes.append(index)
        self._index_name_map[index.name] = index


class Column:
    def __init__(self, identifier: int, name: str, type_: JET_coltyp, record: Record | None = None):
        self.identifier = identifier
        self.name = name
        self.type = type_

        # Set by the table when added, only relevant for fixed value columns
        self._offset = None

        self.record = record

    def __repr__(self) -> str:
        return f"<Column name={self.name!r} identifier={self.identifier:#x} type={self.type} size={self.size}>"

    @property
    def offset(self) -> int:
        return self._offset

    @cached_property
    def is_fixed(self) -> bool:
        return self.identifier <= 127

    @cached_property
    def is_variable(self) -> bool:
        return 127 < self.identifier <= 255

    @cached_property
    def is_tagged(self) -> bool:
        return self.identifier > 255

    @cached_property
    def is_text(self) -> bool:
        return self.type in (JET_coltyp.Text, JET_coltyp.LongText)

    @cached_property
    def is_binary(self) -> bool:
        return self.type in (JET_coltyp.Binary, JET_coltyp.LongBinary)

    @cached_property
    def size(self) -> int:
        if self.record and self.record.get("SpaceUsage"):
            return self.record.get("SpaceUsage")
        return self.ctype.size

    @cached_property
    def default(self) -> Any | None:
        if self.record and self.record.get("DefaultValue"):
            return self.record.get("DefaultValue")
        return None

    @cached_property
    def encoding(self) -> CODEPAGE | None:
        if self.is_text:
            return CODEPAGE(self.record.get("PagesOrLocale")) if self.record else CODEPAGE.ASCII
        return None

    @cached_property
    def ctype(self) -> ColumnType:
        return COLUMN_TYPE_MAP[self.type.value]


class Catalog:
    """Parse and interact with the catalog table.

    The catalog is a special table that contains the metadata for all the other tables in the database.

    Args:
        esedb: An instance of :class:`~dissect.esedb.esedb.EseDB`.
        root_page: The root page of the catalog table.
    """

    CATALOG_COLUMNS = (
        Column(1, "ObjidTable", JET_coltyp.Long),
        Column(2, "Type", JET_coltyp.Short),
        Column(3, "Id", JET_coltyp.Long),
        Column(4, "ColtypOrPgnoFDP", JET_coltyp.Long),
        Column(5, "SpaceUsage", JET_coltyp.Long),
        Column(6, "Flags", JET_coltyp.Long),
        Column(7, "PagesOrLocale", JET_coltyp.Long),
        Column(8, "RootFlag", JET_coltyp.Bit),
        Column(9, "RecordOffset", JET_coltyp.Short),
        Column(10, "LCMapFlags", JET_coltyp.Long),
        Column(11, "KeyMost", JET_coltyp.UnsignedShort),
        Column(12, "LVChunkMax", JET_coltyp.Long),
        Column(128, "Name", JET_coltyp.Text),
        Column(129, "Stats", JET_coltyp.Binary),
        Column(130, "TemplateTable", JET_coltyp.Text),
        Column(131, "DefaultValue", JET_coltyp.Binary),
        Column(132, "KeyFldIDs", JET_coltyp.Binary),
        Column(133, "VarSegMac", JET_coltyp.Binary),
        Column(134, "ConditionalColumns", JET_coltyp.Binary),
        Column(135, "TupleLimits", JET_coltyp.Binary),
        Column(136, "Version", JET_coltyp.Binary),
        Column(137, "SortID", JET_coltyp.Binary),
        Column(256, "CallbackData", JET_coltyp.LongBinary),
        Column(257, "CallbackDependencies", JET_coltyp.LongBinary),
        Column(258, "SeparateLV", JET_coltyp.LongBinary),
        Column(259, "SpaceHints", JET_coltyp.LongBinary),
        Column(260, "SpaceDeferredLVHints", JET_coltyp.LongBinary),
        Column(261, "LocaleName", JET_coltyp.LongBinary),
    )

    def __init__(self, esedb: EseDB, root_page: Page):
        self.esedb = esedb
        self.root_page = root_page
        self.tables = []
        self._table_name_map = {}

        # Create a dummy table with a preset list of columns
        self._ctable = Table(esedb, None, root_page, columns=self.CATALOG_COLUMNS)

        cur_table = None
        for rec in self._ctable.records():
            rtype = rec.get("Type")
            if rtype == SYSOBJ.Table:
                cur_table = Table(self.esedb, rec.get("Name"), rec.get("ColtypOrPgnoFDP"), record=rec)
                self.tables.append(cur_table)
                self._table_name_map[rec.get("Name")] = cur_table

            elif rtype == SYSOBJ.Column:
                column = Column(rec.get("Id"), rec.get("Name"), JET_coltyp(rec.get("ColtypOrPgnoFDP")), record=rec)
                cur_table._add_column(column)

            elif rtype == SYSOBJ.Index:
                index = Index(cur_table, record=rec)
                cur_table._add_index(index)

            elif rtype == SYSOBJ.LongValue:
                cur_table._long_value_record = rec

            elif rtype == SYSOBJ.Callback:
                cur_table._long_callback_record = rec

    def table(self, name: str) -> Table:
        """Retrieve the table with the given name.

        Args:
            name: The table to retrieve.

        Raises:
            KeyError: If no table with that name exists.
        """
        try:
            return self._table_name_map[name]
        except KeyError:
            raise KeyError(f"No table with name: {name}")



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.esedb.btree import BTree
from dissect.esedb.exceptions import NoNeighbourPageError
from dissect.esedb.record import Record

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.esedb.c_esedb import RecordValue
    from dissect.esedb.index import Index
    from dissect.esedb.page import Node


class Cursor:
    """A simple cursor implementation for searching the ESE indexes.

    Args:
        index: The :class:`~dissect.esedb.index.Index` to create the cursor for.
    """

    def __init__(self, index: Index):
        self.index = index
        self.table = index.table
        self.esedb = index.esedb

        self._primary = BTree(self.esedb, index.root)
        self._secondary = None if index.is_primary else BTree(self.esedb, self.table.root)

    def __iter__(self) -> Iterator[Record]:
        while True:
            yield self._record()

            try:
                self._primary.next()
            except NoNeighbourPageError:
                break

    def _node(self) -> Node:
        """Return the node the cursor is currently on. Resolves the secondary index if needed.

        Returns:
            A :class:`~dissect.esedb.page.Node` object of the current node.
        """
        node = self._primary.node()
        if self._secondary:
            self._secondary.reset()
            node = self._secondary.search(node.data.tobytes(), exact=True)
        return node

    def _record(self) -> Record:
        """Return the record the cursor is currently on.

        Returns:
            A :class:`~dissect.esedb.record.Record` object of the current record.
        """
        return Record(self.table, self._node())

    def reset(self) -> None:
        """Reset the internal state."""
        self._primary.reset()
        if self._secondary:
            self._secondary.reset()

    def search(self, **kwargs: RecordValue) -> Record:
        """Search the index for the requested values.

        Searching modifies the cursor state. Searching again will search from the current position.
        Reset the cursor with :meth:`reset` to start from the beginning.

        Args:
            **kwargs: The columns and values to search for.

        Returns:
            A :class:`~dissect.esedb.record.Record` object of the found record.
        """
        key = self.index.make_key(kwargs)
        return self.search_key(key, exact=True)

    def search_key(self, key: bytes, exact: bool = True) -> Record:
        """Search for a record with the given ``key``.

        Args:
            key: The key to search for.
            exact: If ``True``, search for an exact match. If ``False``, sets the cursor on the
                   next record that is greater than or equal to the key.
        """
        self._primary.search(key, exact)
        return self._record()

    def seek(self, **kwargs: RecordValue) -> None:
        """Seek to the record with the given values.

        Args:
            **kwargs: The columns and values to seek to.
        """
        key = self.index.make_key(kwargs)
        self.search_key(key, exact=False)

    def seek_key(self, key: bytes) -> None:
        """Seek to the record with the given ``key``.

        Args:
            key: The key to seek to.
        """
        self._primary.search(key, exact=False)

    def find(self, **kwargs: RecordValue) -> Record | None:
        """Find a record in the index.

        This differs from :meth:`search` in that it will allow additional filtering on non-indexed columns.

        Args:
            **kwargs: The columns and values to search for.
        """
        return next(self.find_all(**kwargs), None)

    def find_all(self, **kwargs: RecordValue) -> Iterator[Record]:
        """Find all records in the index that match the given values.

        This differs from :meth:`search` in that it will allows additional filtering on non-indexed columns.
        If you only search on indexed columns, this will yield all records that match the indexed columns.

        Args:
            **kwargs: The columns and values to search for.
        """
        indexed_columns = {c.name: kwargs.pop(c.name) for c in self.index.columns}
        other_columns = kwargs

        # We need at least an exact match on the indexed columns
        self.search(**indexed_columns)

        current_key = self._primary.node().key

        # Check if we need to move the cursor back to find the first record
        while True:
            if current_key != self._primary.node().key:
                self._primary.next()
                break

            try:
                self._primary.prev()
            except NoNeighbourPageError:
                break

        while True:
            # Entries with the same indexed columns are guaranteed to be adjacent
            if current_key != self._primary.node().key:
                break

            record = self._record()
            for k, v in other_columns.items():
                value = record.get(k)
                # If the record value is a list, we do a check based on the queried value
                if isinstance(value, list):
                    # If the queried value is also a list, we check if they are equal
                    if isinstance(v, list):
                        if value != v:
                            break
                    # Otherwise we check if the queried value is in the record value
                    elif v not in value:
                        break
                else:
                    if value != v:
                        break
            else:
                yield record

            try:
                self._primary.next()
            except NoNeighbourPageError:
                break

    def record(self) -> Record:
        """Return the record the cursor is currently on.

        Returns:
            A :class:`~dissect.esedb.record.Record` object of the current record.
        """
        return self._record()

    def next(self) -> Record:
        """Move the cursor to the next record and return it.

        Can move the cursor to the next page as a side effect.

        Returns:
            A :class:`~dissect.esedb.record.Record` object of the next record.
        """
        try:
            self._primary.next()
        except NoNeighbourPageError:
            raise IndexError("No next record")
        return self._record()

    def prev(self) -> Record:
        """Move the cursor to the previous node and return it.

        Can move the cursor to the previous page as a side effect.

        Returns:
            A :class:`~dissect.esedb.record.Record` object of the previous record.
        """
        try:
            self._primary.prev()
        except NoNeighbourPageError:
            raise IndexError("No previous record")
        return self._record()



import argparse
import datetime
import ipaddress
from collections.abc import Iterator
from pathlib import Path
from typing import BinaryIO

from dissect.util.ts import wintimestamp

from dissect.esedb.c_esedb import RecordValue
from dissect.esedb.esedb import EseDB
from dissect.esedb.table import Table

UalValue = RecordValue | ipaddress.IPv4Address | ipaddress.IPv6Interface | tuple[datetime.datetime]

SKIP_TABLES = [
    "MSysObjects",
    "MSysObjectsShadow",
    "MSysObjids",
    "MSysLocales",
]


class UAL:
    WIN_DATETIME_FIELDS = (
        "CreationTime",
        "FirstSeen",
        "InsertDate",
        "LastAccess",
        "LastSeen",
    )

    def __init__(self, fh: BinaryIO):
        self.esedb = EseDB(fh)

    def get_tables(self) -> list[Table]:
        return self.esedb.tables()

    def get_table_records(self, table_name: str) -> Iterator[dict[str, UalValue]]:
        try:
            table = self.esedb.table(table_name)
        except KeyError:
            return None

        for record in table.records():
            record_data = {}

            last_access_year = None
            day_counts = {}

            for column in table.columns:
                value = record.get(column.name)

                if column.name in self.WIN_DATETIME_FIELDS:
                    value = wintimestamp(value)

                if column.name == "LastAccess":
                    last_access_year = value.year

                if column.name == "Address" and isinstance(value, bytes):
                    value = ipaddress.ip_address(value)
                    value = str(value)

                if column.name.startswith("Day"):
                    day_num = int(column.name[3:])
                    day_counts[day_num] = value
                    continue

                record_data[column.name] = value

            if day_counts:
                if last_access_year:
                    # drop days without counts and convert day number to date
                    day_counts = {
                        convert_day_num_to_date(last_access_year, day_num): count
                        for day_num, count in day_counts.items()
                        if count
                    }
                else:
                    day_counts = {}
                record_data["activity_counts"] = tuple(day_counts.items())

            yield record_data


def convert_day_num_to_date(year: int, day_num: int) -> datetime.datetime:
    return datetime.datetime(year, 1, 1, tzinfo=datetime.timezone.utc) + datetime.timedelta(day_num - 1)


def main() -> None:
    parser = argparse.ArgumentParser(description="dissect.esedb UAL parser")
    parser.add_argument("input", help="UAL database to read")
    args = parser.parse_args()

    with Path(args.input).open("rb") as fh:
        parser = UAL(fh)

        for table in parser.get_tables():
            if table.name in SKIP_TABLES:
                continue

            for record in parser.get_table_records(table.name):
                print(record)


if __name__ == "__main__":
    main()



from unittest.mock import MagicMock

from dissect.esedb.table import Table


def test_find_index() -> None:
    mock_column_id = MagicMock()
    mock_column_id.name = "Id"
    mock_column_bit = MagicMock()
    mock_column_bit.name = "Bit"
    mock_column_unsigned_byte = MagicMock()
    mock_column_unsigned_byte.name = "UnsignedByte"

    mock_idx_id = MagicMock(name="IxId")
    mock_idx_id.is_primary = True
    mock_idx_id.columns = [mock_column_id]
    mock_idx_bit = MagicMock(name="IxBit")
    mock_idx_bit.is_primary = False
    mock_idx_bit.columns = [mock_column_bit]
    mock_idx_multiple = MagicMock(name="IxMultiple")
    mock_idx_multiple.is_primary = False
    mock_idx_multiple.columns = [mock_column_bit, mock_column_unsigned_byte]

    table = Table(MagicMock(), 69, "index", indexes=[mock_idx_id, mock_idx_bit, mock_idx_multiple])

    assert table.find_index(["Id"]) == mock_idx_id
    assert table.find_index(["Bit"]) == mock_idx_bit
    assert table.find_index(["Bit", "UnsignedByte"]) == mock_idx_multiple
    assert table.find_index(["UnsignedByte", "Bit"]) == mock_idx_multiple
    assert table.find_index(["UnsignedByte"]) is None
    assert table.find_index(["Id", "Bit"]) == mock_idx_id
    assert table.find_index(["Bit", "SomethingElse"]) == mock_idx_bit



from __future__ import annotations

from typing import BinaryIO

from dissect.esedb.esedb import EseDB
from dissect.esedb.record import Record


def test_as_dict(basic_db: BinaryIO) -> None:
    db = EseDB(basic_db)
    table = db.table("basic")

    records = list(table.records())
    assert len(records) == 2

    assert [r.as_dict() for r in records] == [
        {
            "Id": 1,
            "Bit": False,
            "UnsignedByte": 213,
            "Short": -1337,
            "Long": -13371337,
            "Currency": 1337133713371337,
            "IEEESingle": 1.0,
            "IEEEDouble": 13371337.13371337,
            "DateTime": 4675210852477960192,
            "UnsignedLong": 13371337,
            "LongLong": -13371337,
            "GUID": "3f360af1-6766-46dc-9af2-0dacf295c2a1",
            "UnsignedShort": 1337,
        },
        {
            "Id": 2,
            "Bit": True,
            "UnsignedByte": 255,
            "Short": 1339,
            "Long": 13391339,
            "Currency": -1339133913391339,
            "IEEESingle": -2.0,
            "IEEEDouble": -13391339.13391339,
            "DateTime": -4537072128574357504,
        },
    ]


def test_comparison(basic_db: BinaryIO) -> None:
    db = EseDB(basic_db)
    table = db.table("basic")

    records = list(table.records())
    assert len(records) == 2

    assert records[0] == records[0]
    assert records[0] != records[1]

    obj = Record(table, records[0]._node)
    assert records[0] == obj
    assert records[0] is not obj

    assert records[0] < records[1]
    assert records[0] <= records[1]
    assert records[0] <= records[0]

    assert set(records) == {records[0], records[1]}
    assert set(records) | {obj} == {records[0], records[1]}


def test_parse_value_encoding(windows_search_db: BinaryIO) -> None:
    """Test if we can parse invalid utf-16-le (Long)Text columns.

    Resources:
        - https://github.com/fox-it/dissect.esedb/pull/48
    """

    db = EseDB(windows_search_db)
    table = db.table("SystemIndex_PropertyStore")

    record = table.search(WorkID=1017)
    auto_summary = record.get("4625-System_Search_AutoSummary")
    assert auto_summary.startswith("Hong Kong SCS AdobeMingStd-Light-Acro-HKscs-B5-H ASCII")
    assert auto_summary.endswith("\\x4c\\xd8")



from __future__ import annotations

from typing import BinaryIO

from dissect.esedb.esedb import EseDB


def test_index(index_db: BinaryIO) -> None:
    db = EseDB(index_db)
    table = db.table("index")

    assert len(table.indexes) == 19

    assert table.indexes[0].name == "IxId"
    assert table.indexes[0].column_ids == [1]
    record = table.indexes[0].search(Id=1)
    assert record.Id == 1

    assert table.indexes[1].name == "IxBit"
    assert table.indexes[1].column_ids == [2]
    record = table.indexes[1].search(Bit=False)
    assert record.Bit is False

    assert table.indexes[2].name == "IxUnsignedByte"
    assert table.indexes[2].column_ids == [3]
    record = table.indexes[2].search(UnsignedByte=213)
    assert record.UnsignedByte == 213

    assert table.indexes[3].name == "IxShort"
    assert table.indexes[3].column_ids == [4]
    record = table.indexes[3].search(Short=-1337)
    assert record.Short == -1337

    assert table.indexes[4].name == "IxLong"
    assert table.indexes[4].column_ids == [5]
    record = table.indexes[4].search(Long=-13371337)
    assert record.Long == -13371337

    assert table.indexes[5].name == "IxCurrenc"
    assert table.indexes[5].column_ids == [6]
    record = table.indexes[5].search(Currency=1337133713371337)
    assert record.Currency == 1337133713371337

    assert table.indexes[6].name == "IxIEEESingle"
    assert table.indexes[6].column_ids == [7]
    record = table.indexes[6].search(IEEESingle=1.0)
    assert record.IEEESingle == 1.0

    assert table.indexes[7].name == "IxIEEEDouble"
    assert table.indexes[7].column_ids == [8]
    record = table.indexes[7].search(IEEEDouble=13371337.13371337)
    assert record.IEEEDouble == 13371337.13371337

    assert table.indexes[8].name == "IxDateTime"
    assert table.indexes[8].column_ids == [9]
    record = table.indexes[8].search(DateTime=4675210852477960192)
    assert record.DateTime == 4675210852477960192

    assert table.indexes[9].name == "IxUnsignedLong"
    assert table.indexes[9].column_ids == [10]
    record = table.indexes[9].search(UnsignedLong=13371337)
    assert record.UnsignedLong == 13371337

    assert table.indexes[10].name == "IxLongLong"
    assert table.indexes[10].column_ids == [11]
    record = table.indexes[10].search(LongLong=-13371337)
    assert record.LongLong == -13371337

    assert table.indexes[11].name == "IxGUID"
    assert table.indexes[11].column_ids == [12]
    record = table.indexes[11].search(GUID="3f360af1-6766-46dc-9af2-0dacf295c2a1")
    assert record.GUID == "3f360af1-6766-46dc-9af2-0dacf295c2a1"

    assert table.indexes[12].name == "IxUnsignedShort"
    assert table.indexes[12].column_ids == [13]
    record = table.indexes[12].search(UnsignedShort=1337)
    assert record.UnsignedShort == 1337

    assert table.indexes[13].name == "IxBinary"
    assert table.indexes[13].column_ids == [128]
    record = table.indexes[13].search(Binary=b"test binary data")
    assert record.Binary == b"test binary data"

    assert table.indexes[14].name == "IxLongBinary"
    assert table.indexes[14].column_ids == [256]
    record = table.indexes[14].search(LongBinary=b"test long binary data " + (b"a" * 1000))
    assert record.LongBinary == b"test long binary data " + (b"a" * 1000)

    assert table.indexes[15].name == "IxASCII"
    assert table.indexes[15].column_ids == [129]
    record = table.indexes[15].search(ASCII="Simple ASCII text")
    assert record.ASCII == "Simple ASCII text"

    assert table.indexes[16].name == "IxUnicode"
    assert table.indexes[16].column_ids == [130]
    record = table.indexes[16].search(Unicode="Simple Unicode text 🦊")
    assert record.Unicode == "Simple Unicode text 🦊"

    assert table.indexes[17].name == "IxLongASCII"
    assert table.indexes[17].column_ids == [257]
    record = table.indexes[17].search(LongASCII="Long ASCII text " + ("a" * 1024))
    assert record.LongASCII == "Long ASCII text " + ("a" * 1024)

    assert table.indexes[18].name == "IxLongUnicode"
    assert table.indexes[18].column_ids == [258]
    record = table.indexes[18].search(LongUnicode="Long Unicode text 🦊 " + ("a" * 1024))
    assert record.LongUnicode == "Long Unicode text 🦊 " + ("a" * 1024)



from __future__ import annotations

import datetime
import struct
import uuid
from typing import NamedTuple

from dissect.cstruct import cstruct

# https://github.com/microsoft/Extensible-Storage-Engine
esedb_def = """
#define MAX_COMPUTERNAME_LENGTH 15

typedef int64 DBTIME;
typedef ULONG OBJID;
typedef ULONG PGNO;
typedef ULONG JET_ENGINEFORMATVERSION;
typedef int64 XECHECKSUM;

enum CODEPAGE {
    UNICODE                 = 1200,
    WESTERN                 = 1252,
    ASCII                   = 20127
};

enum COMPRESSION_SCHEME {
    COMPRESS_NONE           = 0x0,
    COMPRESS_7BITASCII      = 0x1,
    COMPRESS_7BITUNICODE    = 0x2,
    COMPRESS_XPRESS         = 0x3,
    COMPRESS_SCRUB          = 0x4,
    COMPRESS_XPRESS9        = 0x5,
    COMPRESS_XPRESS10       = 0x6,
};

enum JET_coltyp {
    Nil                     = 0,
    Bit                     = 1,
    UnsignedByte            = 2,
    Short                   = 3,
    Long                    = 4,
    Currency                = 5,
    IEEESingle              = 6,
    IEEEDouble              = 7,
    DateTime                = 8,
    Binary                  = 9,
    Text                    = 10,
    LongBinary              = 11,
    LongText                = 12,
    SLV                     = 13,
    UnsignedLong            = 14,
    LongLong                = 15,
    GUID                    = 16,
    UnsignedShort           = 17,
    Max                     = 18,
};

enum SYSOBJ {
    Nil                     = 0,
    Table                   = 1,
    Column                  = 2,
    Index                   = 3,
    LongValue               = 4,
    Callback                = 5,
};

struct LOGTIME {
    BYTE        bSeconds;                           //  0 - 60
    BYTE        bMinutes;                           //  0 - 60
    BYTE        bHours;                             //  0 - 24
    BYTE        bDay;                               //  1 - 31
    BYTE        bMonth;                             //  0 - 11
    BYTE        bYear;                              //  current year - 1900

    BYTE        fTimeIsUTC:1;
    BYTE        bMillisecondsLow:7;

    BYTE        fReserved:1;                        // fOSSnapshot
    BYTE        bMillisecondsHigh:3;
    BYTE        fUnused:4;
};

struct LGPOS {
    USHORT      ib;                                 // must be the last so that lgpos can
    USHORT      isec;                               // index of disksec starting logsec
    LONG        lGeneration;                        // generation of logsec
};

struct SIGNATURE {
    ULONG       ulRandom;                           //  a random number
    LOGTIME     logtimeCreate;                      //  time db created, in logtime format
    CHAR        szComputerName[MAX_COMPUTERNAME_LENGTH + 1];  // where db is created
};

enum BKINFOTYPE : DWORD {
    backupNormal            = 0x0,                  // normal should be 0 for backward compatibility
    backupOSSnapshot,
    backupSnapshot,
    backupSurrogate,                                // should not be persisted in header, but persisted and used by log.
};

struct BKINFO {
    LGPOS       lgposMark;                          //  id for this backup
    LOGTIME     logtimeMark;                        //  timestamp of when le_lgposMark was logged
    ULONG       genLow;                             //  backup set's min lgen
    ULONG       genHigh;                            //  backup set's max lgen
};

struct DBFILEHDR {
    ULONG       ulChecksum;                         //  checksum of the 4k page
    ULONG       ulMagic;                            //  Magic number
    ULONG       ulVersion;                          //  version of DAE the db created (see ulDAEVersion)
    LONG        attrib;                             //  attributes of the db
//  16 bytes

    DBTIME      dbtimeDirtied;                      //  DBTime of this database
//  24 bytes

    SIGNATURE   signDb;                             //  (28 bytes) signature of the db (incl. creation time)
//  52 bytes

    ULONG       dbstate;                            //  consistent/inconsistent state
//  56 bytes

    LGPOS       lgposConsistent;                    //  null if in inconsistent state
    LOGTIME     logtimeConsistent;                  //  null if in inconsistent state
//  72 bytes

    LOGTIME     logtimeAttach;                      //  Last attach time
    LGPOS       lgposAttach;
//  88 bytes

    LOGTIME     logtimeDetach;                      //  Last detach time
    LGPOS       lgposDetach;
//  104 bytes

    ULONG       dbid;                               //  current db attachment
//  108 bytes

    SIGNATURE   signLog;                            //  log signature
//  136 bytes

    BKINFO      bkinfoFullPrev;                     //  Last successful full backup
//  160 bytes

    BKINFO      bkinfoIncPrev;                      //  Last successful Incremental backup
//  184 bytes                                       //  Reset when bkinfoFullPrev is set

    BKINFO      bkinfoFullCur;                      //  current backup. Succeed if a
//  208 bytes                                       //  corresponding pat file generated

    union {
        ULONG   m_ulDbFlags;
        BYTE    m_rgbDbFlags[4];
    };

    OBJID       objidLast;                          //  Object id used so far.

    //  NT version information. This is needed to decide if an index need
    //  be recreated due to sort table changes.

    DWORD       dwMajorVersion;                     //  OS version info
    DWORD       dwMinorVersion;
//  224 bytes

    DWORD       dwBuildNumber;
    LONG        lSPNumber;                          //  use 31 bit only

    ULONG       ulDaeUpdateMajor;                   //  used to track incremental database format updates that
                                                    //  are backward-compatible (see ulDAEUpdateMajorMax)

    ULONG       cbPageSize;                         //  database page size (0 = 4k pages)
//  240 bytes

    ULONG       ulRepairCount;                      //  number of times ErrREPAIRAttachForRepair has been called on this database
    LOGTIME     logtimeRepair;                      //  the date of the last time that repair was run
//  252 bytes

    BYTE        rgbReservedSignSLV[ 28 ];           //  signSLV signature of associated SLV file (obsolete)
//  280 bytes

    DBTIME      dbtimeLastScrub;                    //  last dbtime the database was zeroed out
//  288 bytes

    LOGTIME     logtimeScrub;                       //  the date of the last time that the database was zeroed out
//  296 bytes

    LONG        lGenMinRequired;                    //  the minimum log generation required for replaying the logs. Typically the checkpoint generation
//  300 bytes

    LONG        lGenMaxRequired;                    //  the maximum log generation required for replaying the logs. This is known as the waypoint in BF.
//  304 bytes

    LONG        cpgUpgrade55Format;                 //
    LONG        cpgUpgradeFreePages;                //
    LONG        cpgUpgradeSpaceMapPages;            //
//  316 bytes

    BKINFO      bkinfoSnapshotCur;                  //  Current snapshot.
//  340 bytes

    ULONG       ulCreateVersion;                    //  version of DAE that created db (debugging only)
    ULONG       ulCreateUpdate;
//  348 bytes

    LOGTIME     logtimeGenMaxCreate;                //  creation time of the genMax log file
//  356 bytes

    BKINFOTYPE  bkinfoTypeFullPrev;                 //  Type of Last successful full backup
    BKINFOTYPE  bkinfoTypeIncPrev;                  //  Type of Last successful Incremental backup

//  364 bytes
    ULONG       ulRepairCountOld;                   //  number of times ErrREPAIRAttachForRepair has been called on this database before the last defrag

    ULONG       ulECCFixSuccess;                    //  number of times a one bit error was fixed and resulted in a good page
    LOGTIME     logtimeECCFixSuccess;               //  the date of the last time that a one bit error was fixed and resulted in a good page
    ULONG       ulECCFixSuccessOld;                 //  number of times a one bit error was fixed and resulted in a good page before last repair

    ULONG       ulECCFixFail;                       //  number of times a one bit error was fixed and resulted in a bad page
    LOGTIME     logtimeECCFixFail;                  //  the date of the last time that a one bit error was fixed and resulted in a bad page
    ULONG       ulECCFixFailOld;                    //  number of times a one bit error was fixed and resulted in a bad page before last repair

    ULONG       ulBadChecksum;                      //  number of times a non-correctable ECC/checksum error was found
    LOGTIME     logtimeBadChecksum;                 //  the date of the last time that a non-correctable ECC/checksum error was found
    ULONG       ulBadChecksumOld;                   //  number of times a non-correctable ECC/checksum error was found before last repair

//  416 bytes

    LONG        lGenMaxCommitted;                   //  the last log generation to take active log records for this database.  Not
                                                    //  ensuring replay through this log generation will lose the D in ACID.
//  420 bytes
    BKINFO      bkinfoCopyPrev;                     //  Last successful Copy backup
    BKINFO      bkinfoDiffPrev;                     //  Last successful Differential backup, reset when bkinfoFullPrev is set

//  468 bytes
    BKINFOTYPE  bkinfoTypeCopyPrev;                 //  Type of Last successful Incremental backup
    BKINFOTYPE  bkinfoTypeDiffPrev;                 //  Type of Last successful Differential backup

// 476 bytes
    ULONG       ulIncrementalReseedCount;           //  number of times incremental reseed has been initiated on this database
    LOGTIME     logtimeIncrementalReseed;           //  the date of the last time that incremental reseed was initiated on this database
    ULONG       ulIncrementalReseedCountOld;        //  number of times incremental reseed was initiated on this database before the last defrag

    ULONG       ulPagePatchCount;                   //  number of pages patched in the database as a part of incremental reseed
    LOGTIME     logtimePagePatch;                   //  the date of the last time that a page was patched as a part of incremental reseed
    ULONG       ulPagePatchCountOld;                //  number of pages patched in the database as a part of incremental reseed before the last defrag

// 508 bytes
    QWORD       qwSortVersion;                      // DEPRECATED: In old versions had "default" (?English?) LCID version, in new versions has 0xFFFFFFFFFFFF.

//  516 bytes                                       // checksum during recovery state
    LOGTIME     logtimeDbscanPrev;                  // last checksum finish time (UTC - 1900y)
    LOGTIME     logtimeDbscanStart;                 // start time (UTC - 1900y)
    PGNO        pgnoDbscanHighestContinuous;        // current pgno

//  536 bytes
    LONG        lGenRecovering;                     //  the current log generation that we are doing recovery::redo for.

//  540 bytes

    ULONG       ulExtendCount;
    LOGTIME     logtimeLastExtend;
    ULONG       ulShrinkCount;
    LOGTIME     logtimeLastShrink;

//  564 bytes

    LOGTIME     logtimeLastReAttach;
    LGPOS       lgposLastReAttach;
//  580 bytes

    ULONG       ulTrimCount;
//  584 bytes

    SIGNATURE   signDbHdrFlush;                     //  random signature generated at the time of the last DB header flush
    SIGNATURE   signFlushMapHdrFlush;               //  random signature generated at the time of the last FM header flush
//  640 bytes

    LONG        lGenMinConsistent;                  //  the minimum log generation required to bring the database to a clean state
                                                    //  it might be different from lGenMinRequired, which also encompasses flush map consistency
//  644 bytes

    ULONG       ulDaeUpdateMinor;                   //  used to track incremental database format updates that
                                                    //  are forwards-compatible (see ulDAEUpdateMinorMax)
    JET_ENGINEFORMATVERSION efvMaxBinAttachDiagnostic;  //  Max version of engine binary that has attached this database.
                                                        //      NOTE: This is NOT necessarily the format the DB creates/should maintain if JET_paramEnableFormatVersion is set.
                                                        //      NOTE: ALSO this is NOT propagated via redo, intentionally to avoid dependency on this.  Use le_ulDaeUpdateMinor instead.
//  652 bytes

    PGNO        pgnoDbscanHighest;                  // highest pgno - used to avoid re-scanning pages
//  656 bytes

    LGPOS       lgposLastResize;                    // lgpos of the last database resize committed to the database.
                                                    // This is unset on older builds and it requires JET_efvLgposLastResize to start being populated.
                                                    // A database attachment sets this to le_lgposAttach, even though there isn't a real resize, so it can be
                                                    // be viewed as "resize checkpoint".
                                                    // A database extension or shrinkage updates this to the LGPOS of the operation.
                                                    // A clean databse detachment sets this to le_lgposConsistent.
                                                    // Incremental reseed may set this back if the first divergent log is lower than the current le_lgposLastResize.
//  664 bytes

    BYTE        rgbReserved[3];                     // keeping the le_filetype in the same
                                                    // place as log file header and checkpoint
                                                    // file header
                                                    // Remember: when updating logfile, checkpoint or
                                                    // flush map file headers, please check here
//  667 bytes

    //  WARNING: MUST be placed at this offset for
    //  uniformity with db/log headers
    ULONG       filetype;                           //  JET_filetypeDatabase or JET_filetypeStreamingFile
//  671 bytes

    BYTE        rgbReserved2[1];                    // For alignment
//  672 bytes

    LOGTIME     logtimeGenMaxRequired;
//  680 bytes

    LONG        lGenPreRedoMinConsistent;           //  the minimum log generation required to bring the database to a clean state (value overriden during the start of redo, 0 if not overriden).
    LONG        lGenPreRedoMinRequired;             //  the minimum log generation required for replaying the logs (value overriden during the start of redo, 0 if not overriden).
//  688 bytes

    SIGNATURE   signRBSHdrFlush;                    //  random signature generated at the time of the last RBS header flush
//  716 bytes

    ULONG       ulRevertCount;                      //  number of times revert has been initiated on this database using the revert snapshots.
    LOGTIME     logtimeRevertFrom;                  //  the date of the last time that revert was initiated on this database using the revert snapshots.
    LOGTIME     logtimeRevertTo;                    //  the date of the last time we were reverting this database to.
    ULONG       ulRevertPageCount;                  //  number of pages reverted on the database by the recent revert operation.
    LGPOS       lgposCommitBeforeRevert;            //  lgpos of the last commit on the database before a revert was done and requires JET_efvApplyRevertSnapshot to start being populated.
                                                    //  This will be set only after a revert is completed and will be used to ignore JET_errDbTimeTooOld on passive copies.
                                                    //  This is because as part of revert any new page reverts will cause the page to be zeroed out and thereby might be behind the active's dbtime for the page.
};

flag PAGE_FLAG : uint32 {
    // where we are in the BTree
    Root                    = 0x00000001,           // fPageRoot
    Leaf                    = 0x00000002,           // fPageLeaf
    ParentOfLeaf            = 0x00000004,           // fPageParentOfLeaf
    // special flags
    Empty                   = 0x00000008,           // fPageEmpty
    Repair                  = 0x00000010,           // fPageRepair
    // what type of tree we are in
    Primary                 = 0x00000000,           // fPagePrimary
    SpaceTree               = 0x00000020,           // fPageSpaceTree
    Index                   = 0x00000040,           // fPageIndex
    LongValue               = 0x00000080,           // fPageLongValue
    // type of BTree key validation
    NonUniqueKeys           = 0x00000400,           // fPageNonUniqueKeys
    // upgrade info
    NewRecordFormat         = 0x00000800,           // fPageNewRecordFormat
    NewChecksumFormat       = 0x00002000,           // fPageNewChecksumFormat
    Scrubbed                = 0x00004000,           // fPageScrubbed
};

struct PGHDR {
    XECHECKSUM  checksum;
    DBTIME      dbtimeDirtied;
    PGNO        pgnoPrev;
    PGNO        pgnoNext;
    OBJID       objidFDP;
    USHORT      cbFree;
    USHORT      cbUncommittedFree;
    USHORT      ibMicFree;
    USHORT      itagState;
    PAGE_FLAG   fFlags;
};

struct PGHDR2 {
    // PGHDR       pghdr;
    XECHECKSUM  rgChecksum[3];
    PGNO        pgno;
    CHAR        rgbReserved[12];
};

flag TAG_FLAG : uint16 {
    Version                 = 0x01,                 // fNDVersion
    Deleted                 = 0x02,                 // fNDDeleted
    Compressed              = 0x04,                 // fNDCompressed
};

struct TAG {
    USHORT      cb_;                                // size
    USHORT      ib_;                                // offset
};

typedef WORD RECOFFSET;

struct RECHDR {
    BYTE        fidFixedLastInRec;
    BYTE        fidVarLastInRec;
    RECOFFSET   ibEndOfFixedData;
};

flag TAGFLD_HEADER : uint8 {
    Invalid                 = 0x00,                 // Not a real flag
    LongValue               = 0x01,                 // fLongValue, is the column type JET_coltypLongText or JET_coltypLongBinary
    Compressed              = 0x02,                 // fCompressed
    Separated               = 0x04,                 // fSeparated, is the data stored in a long value page
    MultiValues             = 0x08,                 // fMultiValues
    TwoValues               = 0x10,                 // fTwoValues
    Null                    = 0x20,                 // fNull
    Encrypted               = 0x40,                 // fEncrypted
};

flag JET_bitIndex : uint32 {
    Unique                  = 0x00000001,
    Primary                 = 0x00000002,
    DisallowNull            = 0x00000004,
    IgnoreNull              = 0x00000008,
    Clustered40             = 0x00000010,           // for backward compatibility
    IgnoreAnyNull           = 0x00000020,
    IgnoreFirstNull         = 0x00000040,
    LazyFlush               = 0x00000080,
    Empty                   = 0x00000100,           // don't attempt to build index, because all entries would evaluate to NULL (MUST also specify JET_bitIgnoreAnyNull)
    Unversioned             = 0x00000200,
    SortNullsHigh           = 0x00000400,           // NULL sorts after data for all columns in the index
    Unicode                 = 0x00000800,
    Tuples                  = 0x00001000,           // index on substring tuples (text columns only)
    TupleLimits             = 0x00002000,           // cbVarSegMac field of JET_INDEXCREATE actually points to a JET_TUPLELIMITS struct to allow custom tuple index limits (implies JET_bitIndexTuples)
    CrossProduct            = 0x00004000,           // index over multiple multi-valued columns has full cross product
    KeyMost                 = 0x00008000,           // custom index key size set instead of default of 255 bytes
    DisallowTruncation      = 0x00010000,           // fail update rather than truncate index keys
    NestedTable             = 0x00020000,           // index over multiple multi-valued columns but only with values of same itagSequence
    DotNetGuid              = 0x00040000,           // index over GUID column according to .Net GUID sort order
    ImmutableStructure      = 0x00080000,           // Do not write to the input structures during a JetCreateIndexN call.
};

flag IDBFLAG : uint16 {
    Unique                  = 0x0001,               // Duplicate keys not allowed
    AllowAllNulls           = 0x0002,               // Make entries for NULL keys (all segments are null)
    AllowFirstNull          = 0x0004,               // First index column NULL allowed in index
    AllowSomeNulls          = 0x0008,               // Make entries for keys with some null segments
    NoNullSeg               = 0x0010,               // Don't allow a NULL key segment
    Primary                 = 0x0020,               // Index is the primary index
    LocaleSet               = 0x0040,               // Index locale information (locale name) is set (JET_bitIndexUnicode was specified).
    Multivalued             = 0x0080,               // Has a multivalued segment
    TemplateIndex           = 0x0100,               // Index of a template table
    DerivedIndex            = 0x0200,               // Index derived from template table
                                                    //   Note that this flag is persisted, but
                                                    //   never used in an in-memory IDB, because
                                                    //   we use the template index IDB instead.
    LocalizedText           = 0x0400,               // Has a unicode text column? (code page is 1200)
    SortNullsHigh           = 0x0800,               // NULL sorts after data
    // Jan 2012: MSU is being removed. fidbUnicodeFixupOn should no longer be referenced.
    UnicodeFixupOn_Deprecated   = 0x1000,           // Track entries with undefined Unicode codepoints
    CrossProduct            = 0x2000,               // all combinations of multi-valued columns are indexed
    DisallowTruncation      = 0x4000,               // fail update rather than allow key truncation
    NestedTable             = 0x8000,               // combinations of multi-valued columns of same itagSequence are indexed
};

flag IDXFLAG : uint16 {
    ExtendedColumns         = 0x0001,               // IDXSEGs are comprised of JET_COLUMNIDs, not FIDs
    DotNetGuid              = 0x0002,               // GUIDs sort according to .Net rules
};
"""  # noqa E501

c_esedb = cstruct().load(esedb_def)

ulDAEMagic = 0x89ABCDEF
pgnoFDPMSO = 4
pgnoFDPMSO_NameIndex = 7
pgnoFDPMSO_RootObjectIndex = 10
pgnoFDPMSOShadow = 24

JET_coltyp = c_esedb.JET_coltyp
JET_bitIndex = c_esedb.JET_bitIndex
SYSOBJ = c_esedb.SYSOBJ
PAGE_FLAG = c_esedb.PAGE_FLAG
TAG_FLAG = c_esedb.TAG_FLAG
TAGFLD_HEADER = c_esedb.TAGFLD_HEADER
CODEPAGE = c_esedb.CODEPAGE
COMPRESSION_SCHEME = c_esedb.COMPRESSION_SCHEME
IDBFLAG = c_esedb.IDBFLAG
IDXFLAG = c_esedb.IDXFLAG

CODEPAGE_MAP = {
    CODEPAGE.UNICODE: "utf-16-le",
    CODEPAGE.WESTERN: "cp1252",
    CODEPAGE.ASCII: "ascii",
}

RecordValue = int | float | str | bytes | datetime.datetime | None


def decode_bit(buf: bytes) -> bool:
    """Decode a bit into a boolean.

    Args:
        buf: The buffer to decode from.
    """
    return c_esedb.uint8(buf) == 0xFF


def decode_text(buf: bytes, encoding: CODEPAGE, errors: str | None = "backslashreplace") -> str:
    """Decode text with the appropriate encoding.

    Args:
        buf: The buffer to decode from.
    """
    buf = bytes(buf)

    if encoding == CODEPAGE.UNICODE and len(buf) % 2:
        buf += b"\x00"

    return buf.decode(CODEPAGE_MAP[encoding], errors=errors).rstrip("\x00")


def decode_guid(buf: bytes) -> str:
    """Decode a GUID.

    Args:
        buf: The buffer to decode from.
    """
    return str(uuid.UUID(bytes_le=bytes(buf)))


def checksum_xor(data: bytes, initial: int = 0x89ABCDEF) -> int:
    digest = initial
    for val in struct.unpack(f"<{len(data) // 4}I", data):
        digest ^= val

    return digest


class ColumnType(NamedTuple):
    value: int
    name: str
    size: int | None
    parse: callable | None


COLUMN_TYPES = [
    ColumnType(JET_coltyp.Nil, "NULL", 0, None),
    ColumnType(JET_coltyp.Bit, "Boolean", 1, decode_bit),
    ColumnType(JET_coltyp.UnsignedByte, "Unsigned byte", 1, c_esedb.uint8),
    ColumnType(JET_coltyp.Short, "Signed short", 2, c_esedb.int16),
    ColumnType(JET_coltyp.Long, "Signed long", 4, c_esedb.int32),
    ColumnType(JET_coltyp.Currency, "Currency", 8, c_esedb.int64),
    ColumnType(JET_coltyp.IEEESingle, "Single precision FP", 4, c_esedb.float),
    ColumnType(JET_coltyp.IEEEDouble, "Double precision FP", 8, c_esedb.double),
    # Parse DateTime as an int64 because the actual parsing of the value can differ between databases
    # E.g. by default it's supposed to be an OA date, but the UAL stores it as a regular Windows timestamp
    ColumnType(JET_coltyp.DateTime, "DateTime", 8, c_esedb.int64),
    ColumnType(JET_coltyp.Binary, "Binary", None, bytes),
    ColumnType(JET_coltyp.Text, "Text", None, decode_text),
    ColumnType(JET_coltyp.LongBinary, "Long Binary", None, bytes),
    ColumnType(JET_coltyp.LongText, "Long Text", None, decode_text),
    ColumnType(JET_coltyp.SLV, "Super Long Value", None, None),
    ColumnType(JET_coltyp.UnsignedLong, "Unsigned long", 4, c_esedb.uint32),
    ColumnType(JET_coltyp.LongLong, "Signed Long long", 8, c_esedb.int64),
    ColumnType(JET_coltyp.GUID, "GUID", 16, decode_guid),
    ColumnType(JET_coltyp.UnsignedShort, "Unsigned short", 2, c_esedb.uint16),
    ColumnType(JET_coltyp.Max, "Max", None, None),
]
COLUMN_TYPE_MAP = {t.value.value: t for t in COLUMN_TYPES}



from __future__ import annotations

import runpy
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any

from dissect.esedb import EseDB

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.esedb.c_esedb import RecordValue
    from dissect.esedb.record import Record

try:
    from impacket import ese
except ImportError:
    print("This utility monkeypatches ESENT_DB from impacket, but impacket is not installed", file=sys.stderr)
    exit(1)


class ESENT_DB:
    def __init__(self, fileName: Any, pageSize: int = 8192, isRemote: bool = False) -> None:
        # I don't know what ``isRemote`` actually does so mimick what Impacket does
        if isRemote:
            self.fh = fileName
            self.fh.open()
        else:
            self.fh = Path(fileName).open("rb")  # noqa: SIM115
        self.db = EseDB(self.fh, impacket_compat=True)

    def openTable(self, tableName: str) -> Iterator[Record]:
        # Impacket only ever asks for the "next" record, so our cursor can simply be the record iterator
        return self.db.table(tableName).records()

    def getNextRow(self, cursor: Iterator[Record], filter_tables: list[str] | None = None) -> RecordWrapper | None:
        # Impacket uses a list to filter column names to make parsing a bit more efficient, but our parsing already
        # skips columns you don't request the value for
        try:
            # Impacket treats column names as bytes so we need to wrap all records to translate column lookup
            return RecordWrapper(next(cursor))
        except StopIteration:
            return None

    def close(self) -> None:
        self.fh.close()


class RecordWrapper:
    def __init__(self, record: Record):
        self._record = record

    def __getitem__(self, attr: bytes) -> RecordValue:
        # Impacket treats column names as bytes so decode to string
        return self._record.get(attr.decode())


ese.ESENT_DB = ESENT_DB


if __name__ == "__main__":
    if len(sys.argv) == 1:
        print(
            "Usage: python -m dissect.esedb.tools.impacket /path/to/impacket/secretsdump.py <arguments>",
            file=sys.stderr,
        )
        exit(1)

    del sys.argv[0]
    main_globals = sys.modules["__main__"].__dict__
    runpy.run_path(sys.argv[0], main_globals, "__main__")



from __future__ import annotations

from typing import BinaryIO

from dissect.esedb.tools.ual import UAL


def test_ual(ual_db: BinaryIO) -> None:
    db = UAL(ual_db)

    assert len(list(db.get_table_records("CLIENTS"))) == 19
    assert len(list(db.get_table_records("ROLE_ACCESS"))) == 3
    assert len(list(db.get_table_records("VIRTUALMACHINES"))) == 0
    assert len(list(db.get_table_records("DNS"))) == 12
    assert len(list(db.get_table_records("SYSTEM_IDENTITY"))) == 0



from typing import BinaryIO

from dissect.esedb.esedb import EseDB


def test_cursor(basic_db: BinaryIO) -> None:
    db = EseDB(basic_db)
    table = db.table("basic")
    idx = table.index("IxId")

    cursor = idx.cursor()
    record = cursor.search(Id=1)
    assert record.Id == 1
    record = cursor.next()
    assert record.Id == 2
    record = cursor.prev()
    assert record.Id == 1
    assert record.Id == cursor.record().Id


def test_cursor_iterator(basic_db: BinaryIO) -> None:
    db = EseDB(basic_db)
    table = db.table("basic")
    idx = table.index("IxId")

    cursor = idx.cursor()
    records = list(cursor)
    assert len(records) == 2
    assert records[0].Id == 1
    assert records[1].Id == 2


def test_cursor_search(ual_db: BinaryIO) -> None:
    db = EseDB(ual_db)
    table = db.table("CLIENTS")
    idx = table.index("Username_RoleGuid_TenantId_index")

    cursor = idx.cursor()
    records = list(
        cursor.find_all(
            AuthenticatedUserName="blackclover\\administrator",
            RoleGuid="ad495fc3-0eaa-413d-ba7d-8b13fa7ec598",
            TenantId="2417e4c3-5467-40c5-809b-12b59a86c102",
        )
    )

    assert len(records) == 5

    cursor.reset()
    records = list(
        cursor.find_all(
            AuthenticatedUserName="blackclover\\administrator",
            RoleGuid="ad495fc3-0eaa-413d-ba7d-8b13fa7ec598",
            TenantId="2417e4c3-5467-40c5-809b-12b59a86c102",
            Day204=4,
        )
    )

    assert len(records) == 1



from __future__ import annotations

import struct
from functools import cached_property
from typing import TYPE_CHECKING

from dissect.esedb.c_esedb import PAGE_FLAG, TAG_FLAG, c_esedb

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.esedb.esedb import EseDB


class Page:
    """Represents a logical page of an ESE database.

    Args:
        esedb: An instance of :class:`~dissect.esedb.esedb.EseDB`.
        num: The logical page number.
        buf: The physical page data.
    """

    def __init__(self, esedb: EseDB, num: int, buf: bytes):
        self.esedb = esedb
        self.num = num
        self.buf = memoryview(buf)

        data_start = len(c_esedb.PGHDR)
        self.header = c_esedb.PGHDR(self.buf)
        self.header2 = None
        if not self.is_small_page:
            # Large pages have some additional header fields
            self.header2 = c_esedb.PGHDR2(self.buf[data_start:])
            data_start += len(c_esedb.PGHDR2)

        self.flags = self.header.fFlags
        self.previous_page = self.header.pgnoPrev
        self.next_page = self.header.pgnoNext

        data_end = self.header.ibMicFree
        self.data = self.buf[data_start : data_start + data_end]

        # On newer versions, itagMicFree was renamed to itagState and
        # now includes a reserved tag count in the upper 4 bits
        self.tag_reserved = (self.header.itagState >> 12) or 1
        self.tag_count = self.header.itagState & 0x0FFF
        self.node_count = self.tag_count - self.tag_reserved
        self._node_cls = LeafNode if self.is_leaf else BranchNode
        self._node_cache = {}

    @cached_property
    def is_small_page(self) -> bool:
        return self.esedb.has_small_pages

    @cached_property
    def is_root(self) -> bool:
        return bool(self.flags & PAGE_FLAG.Root)

    @cached_property
    def is_leaf(self) -> bool:
        return bool(self.flags & PAGE_FLAG.Leaf)

    @cached_property
    def is_parent(self) -> bool:
        return bool(self.flags & PAGE_FLAG.ParentOfLeaf)

    @cached_property
    def is_empty(self) -> bool:
        return bool(self.flags & PAGE_FLAG.Empty)

    @cached_property
    def is_space_tree(self) -> bool:
        return bool(self.flags & PAGE_FLAG.SpaceTree)

    @cached_property
    def is_index(self) -> bool:
        return bool(self.flags & PAGE_FLAG.Index)

    @cached_property
    def is_long_value(self) -> bool:
        return bool(self.flags & PAGE_FLAG.LongValue)

    @cached_property
    def is_branch(self) -> bool:
        return not self.is_leaf

    @cached_property
    def key_prefix(self) -> bytes | None:
        if not self.is_root:
            return bytes(self.tag(0).data)

        return None

    def tag(self, num: int) -> Tag:
        """Retrieve a tag by index.

        Args:
            num: The tag number to retrieve.

        Raises:
            IndexError: If the tag number is out of bounds.
        """
        if num < 0 or num > self.tag_count - 1:
            raise IndexError(f"Tag number exceeds boundaries: 0-{self.tag_count - 1}")

        return Tag(self, num)

    def tags(self) -> Iterator[Tag]:
        """Yield all tags."""
        for i in range(1, self.tag_count):
            yield self.tag(i)

    def node(self, num: int) -> BranchNode | LeafNode:
        """Retrieve a node by index.

        Nodes are just tags, but indexed from the first tag.

        Args:
            num: The node number to retrieve.

        Raises:
            IndexError: If the node number is out of bounds.
        """
        if num < 0 or num > self.node_count - 1:
            raise IndexError(f"Node number exceeds boundaries: 0-{self.node_count - 1}")

        if num not in self._node_cache:
            self._node_cache[num] = self._node_cls(self.tag(num + 1))

        return self._node_cache[num]

    def nodes(self) -> Iterator[BranchNode | LeafNode]:
        """Yield all nodes."""
        for i in range(self.node_count):
            yield self.node(i)

    def iter_leaf_nodes(self) -> Iterator[LeafNode]:
        """Walk the page tree and yield leaf nodes.

        Two methods can be used, one is to travel down to the first leaf, and keep reading ``next_page``'s,
        the other is to traverse the tree branches.

        Impacket uses the first method, but gets caught in an infinite loop on some dirty databases.
        Traversing the branches seems safer, at the risk of missing a couple (possibly corrupt) pages.

        For this reason, we actually explicitly check if the last page we parse has a ``next_page`` attribute,
        and also parse that. This methods seems to work so far.
        """
        esedb = self.esedb
        leaf = None

        for node in self.nodes():
            if self.is_leaf:
                yield node
            else:
                child = esedb.page(node.child)
                for leaf in child.iter_leaf_nodes():
                    yield leaf

        if self.is_root and leaf and leaf.tag.page.next_page:
            yield from esedb.page(leaf.tag.page.next_page).iter_leaf_nodes()

    def __repr__(self) -> str:
        return f"<Page num={self.num:d}>"


class Tag:
    """A tag is the "physical" data entry of a page.

    Args:
        page: The :class:`Page` this tag is in.
        num: The tag number to parse.
    """

    __slots__ = ("data", "flags", "num", "offset", "page", "size", "tag")

    def __init__(self, page: Page, num: int):
        self.page = page
        self.num = num

        tag_offset = len(page.buf) + ((num + 1) * -4)
        tag_buf = page.buf[tag_offset : tag_offset + 4]

        self.tag = c_esedb.TAG(tag_buf)

        mask = 0x1FFF if page.is_small_page else 0x7FFF
        self.size = self.tag.cb_ & mask
        self.offset = self.tag.ib_ & mask

        if self.size == 0 and self.num != 0:
            raise ValueError("Invalid TAG data, corrupt database?")

        self.data = self.page.data[self.offset : self.offset + self.size]

        flags = 0
        if page.is_small_page:
            # Small pages have the flag in the tag
            flags = self.tag.ib_ >> 13
        elif len(self.data) >= 2:
            # Large pages have the flag in the first USHORT
            # Also in the 3 MSB, just like the small page, but we know it'll be little endian so do a shortcut on
            # the second byte.
            flags = self.data[1] >> 5

        self.flags = TAG_FLAG(flags)

    def __repr__(self) -> str:
        return f"<Tag offset={self.offset:#x} size={self.size:#x}>"


class Node:
    """A node is the "logical" data entry of a page.

    Args:
        tag: The :class:`Tag` to parse a node from.
    """

    __slots__ = ("data", "key", "key_prefix", "key_suffix", "num", "tag")

    def __init__(self, tag: Tag):
        self.tag = tag
        self.num = tag.num - 1

        buf = tag.data
        offset = 0

        key_prefix = b""
        key_prefix_size = None

        # Large pages have the tag flags encoded in the 3 MSB of the first word, so we have to mask the first 13 bits
        # See also the Tag class
        if len(buf) >= offset + 2 and tag.flags & TAG_FLAG.Compressed:
            key_prefix_size = struct.unpack("<H", buf[:2])[0] & 0x1FFF
            key_prefix = self.tag.page.key_prefix[:key_prefix_size].ljust(key_prefix_size, b"\x00")
            offset += 2

        key_suffix = b""
        key_suffix_size = None

        if len(buf) >= offset + 2:
            key_suffix_size = struct.unpack("<H", buf[offset : offset + 2])[0] & 0x1FFF
            offset += 2
            key_suffix = buf[offset : offset + key_suffix_size]
            offset += key_suffix_size

        self.key = key_prefix + key_suffix
        self.key_prefix = key_prefix
        self.key_suffix = key_suffix
        self.data = buf[offset:]


class LeafNode(Node):
    """Special leaf node."""

    def __repr__(self) -> str:
        return f"<LeafNode key={self.key}>"


class BranchNode(Node):
    """Special branch node. Parses the child page information."""

    __slots__ = ("child",)

    def __init__(self, tag: Tag):
        super().__init__(tag)
        self.child = struct.unpack("<I", self.data[:4])[0]

    def __repr__(self) -> str:
        return f"<BranchNode key={self.key} child={self.child}>"
